#!/bin/sh

#
# This script makes easier using the NS3D code on Zahir (Cluster Power4) at IDRIS
# It performs all the necessary steps:
#     - copying the necessary files (input data, sources...) into the execution directory
#     - compilation
#     - execution
#     - save of the results in the save directory
#
# It must be run from the computer connected to the exterior "griffu"


#____________________________________________________________
#
#    PATHS/NAMES
#____________________________________________________________
#

# ------------------------------
# Directories/files used
# ------------------------------

# directory where the results are saved on Gaya (a sub-directory is automatically created for each run)
dir_save_gaya="/u/rech/bll/rbll200/NS3D/"

# directory where the source files are
dir_source="../source"

# optional paths to the input data files (previous results data file to restart a run, initial velocity, base flow)
velocity_init=""
base2d_init=""
velocity_init_gaya="/u/rech/bll/rbll200/NS3D/velo_rho_vort.t=0007.500_1024x1408x192_L=1.5x2.1x0.285"

# name of the simulation, useful to determine the name of the working/save folders
run_name="contrarotating"

# Number of parallel OpenMP processes (0 means non-OpenMP code)
nb_openmp_processes=0

# Number of parallel MPI processes (0 means non-MPI code)
nb_mpi_processes=64

# Number of recursive submission of the script. It is possible to perform several runs in a row
# using the final velocity of the previous run as initial velocity (0 or 1 means single run)
nb_recursive_run=10

# Mode of execution ("batch" or "interactive")
mode="batch"

# In batch mode, it is necessary to indicate some additional parameters
#   - name of the job (optional)
#   - cpu time in seconds. The unit must not be indicated.
#   - data memory used and the unit used (Mb, Gb)
#   - stack memory used and the unit used (Mb, Gb)
name_ll="ns3d"
time_cpu_ll="72000"
memory_data_ll="1.3Gb"
memory_stack_ll="0.2Gb"

#____________________________________________________________
#
#    COMPILATION PARAMETERS (config.h)
#____________________________________________________________
#

mkdir -p temp
cat > temp/config.h_pid$$ << EOF
! *******************************************
!    NS3D: compilation parameters
! *******************************************

! This file is used by the pre-processor during the compilation step.

! Avoid multiple inclusions of this file. Do not change!
#ifndef CONFIG_H
#define CONFIG_H

! Number of colocation points
#define DIMX 1024
#define DIMY 1408
#define DIMZ 192

! Padding along each direction
#define PPADKX 0
#define PPADKY 0
#define PPADKZ 0

! The following time schemes are available:
!     - AB2: Adams-Bashforth of order 2
!     - RK2: Runge and Kutta of order 2
!     - RK3: Runge and Kutta of order 3
!     - RK4: Runge and Kutta of order 4
#define RK4

! We can simulate a single complex kz mode for a perturbative and linearized run.
! Warning: for correct results, the eigenmode growhtrate must be purely real.
#undef SINGLE_MODE

! we estimate a translation velocity of a basic state
#undef calcul_ut

#endif
EOF



#____________________________________________________________
#
#    SIMULATION PARAMETERS (data.in)
#____________________________________________________________
#

cat > temp/data.in_pid$$ << EOF
********************************************
*    NS3D: simulation parameters           *
********************************************

This file is read at every simulation start.

*** discretization variables ***
lx___________________________   1.5
ly___________________________   2.1
lz___________________________   0.285
dt___________________________   0.001
begin________________________   0.
itmax________________________   500
de_aliasing__________________   T
    squared:1_or_elliptic:2__   1
    radius_truncation_x______   0.95
    radius_truncation_y______   0.95
    radius_truncation_z______   0.95

*** physical variables ***
viscosity____________________   3.E-5
stratified___________________   T
    brunt_vaisala_frequency__   24.
    schmidt_number___________   1.
2omega_______________________   0.

*** Type of simulation ***
perturbative_________________   F
    linear___________________   T

*** Base state (only perturbation run) ***
base2D_type__________________   null

*** Initial velocity ***
velo_type____________________   file
white_noise__________________   0.

*** Output ***
output1_period_______________   100
output2_period_______________   250
output3_period_______________   0


*** additionnal variables... ***

* Stuart vortices *
concentration_stuart_________   0.25

* File gaussian vortices *
file_nb_vortices_____________   2
Vortex 1
    position_x_______________   2.1
    position_y_______________   2.35
    circulation______________   2.
    core_radius______________   0.2
Vortex 2
    position_x_______________   2.1
    position_y_______________   1.85
    circulation______________   -2.
    core_radius______________   0.2

* Random gaussian vortices *
rnd_nb_vortices______________   10
rnd_min_distance_____________   1.
rnd_mean_gamma_______________   6.28
rnd_std_gamma________________   0.
rnd_mean_radius______________   1.
rnd_std_radius_______________   0.

EOF


#____________________________________________________________
#
#     DO NOT MODIFY BELOW!
#____________________________________________________________
#

printf "\n********** START OF THE NS3D SCRIPT *********\n"


#_____________________________________________________
#
#    PRELIMINARY TASKS
#_____________________________________________________
#

# useful to personalize the names of the used directories
date_hour=`date +"%d-%m-%y_%Hh%M"`
pid_job=$$

# we read some simulation parameters in the configuration files
NX=`grep "#define DIMX" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMX/ {print $3}'`
NY=`grep "#define DIMY" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMY/ {print $3}'`
NZ=`grep "#define DIMZ" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMZ/ {print $3}'`
LX=`awk '/^lx/ {print $2}' temp/data.in_pid$$`
LY=`awk '/^ly/ {print $2}' temp/data.in_pid$$`
LZ=`awk '/^lz/ {print $2}' temp/data.in_pid$$`
Dt=`awk '/^dt/ {print $2}' temp/data.in_pid$$`
nu=`awk '/^viscosity/ {print $2}' temp/data.in_pid$$`
stratifie=`awk '/^stratified/ {print $2}' temp/data.in_pid$$`
brunt=`awk '/brunt_vaisala_frequency/ {print $2}' temp/data.in_pid$$`

if [ "$stratifie" = "F" ]; then
    brunt=0
fi

dir_name="${run_name}_${NX}x${NY}x${NZ}_L=${LX}x${LY}x${LZ}_Dt=${Dt}_nu=${nu}_N=${brunt}"


# we determine the Makefile to use depending on the number of MPI and OpenMP processes
if [ $nb_openmp_processes -eq 0 -a $nb_mpi_processes -eq 0 ] ; then
    makefile_file=${dir_source}/Makefile_examples/Makefile_IBM_SP_essl
elif [ $nb_openmp_processes -gt 0 -a $nb_mpi_processes -eq 0 ] ; then
    makefile_file=${dir_source}/Makefile_examples/Makefile_IBM_SP_essl_OpenMP
elif [ $nb_openmp_processes -eq 0 -a $nb_mpi_processes -gt 0 ] ; then
    makefile_file=${dir_source}/Makefile_examples/Makefile_IBM_SP_fftw_MPI
else
    makefile_file=${dir_source}/Makefile_examples/Makefile_IBM_SP_fftw_MPI_OpenMP
fi

printf "Makefile: ${makefile_file}\n"


#_____________________________________________________
#
#    MODIFICATION OF THE EXECUTION SCRIPT "job_zahir_exec"
#_____________________________________________________
#

# in batch mode, no redirection of the output messages
if [ "$mode" = "batch" ] ; then
    arg1=""
else
    arg1=$1
fi

if [ $nb_recursive_run -gt 1 ] ; then
    nb_recursive_run_left=`expr $nb_recursive_run - 1`
else
    nb_recursive_run_left=0
fi

sed -e s\%^dir_save_gaya=.*%dir_save_gaya=$dir_save_gaya% \
    -e "s/^nb_openmp_processes=.*/nb_openmp_processes=$nb_openmp_processes/" \
    -e "s/^nb_mpi_processes=.*/nb_mpi_processes=$nb_mpi_processes/" \
    -e "s/^arg1=.*/arg1="$1"/" \
    -e "s/^date_hour=.*/date_hour=$date_hour/" \
    -e "s/^pid_job=.*/pid_job=$pid_job/" \
    -e "s/^dir_name=.*/dir_name=$dir_name/" \
    -e "s/^nb_recursive_run_left=.*/nb_recursive_run_left=$nb_recursive_run_left/" \
    job_zahir_exec.in > temp/job_zahir_exec_pid$$

sed -e "s/-DMPI=[0-9]*/-DMPI=$nb_mpi_processes/" \
    -e "s/-DFFTW_MPI_OPENMP=[0-9]*/-DFFTW_MPI_OPENMP=$nb_openmp_processes/" \
    $makefile_file > temp/Makefile_pid$$

# in batch mode, we write the submission directives in the script header
if [ "$mode" = "batch" ] ; then

    if [ $nb_openmp_processes -eq 0 -a $nb_mpi_processes -eq 0 ] ; then
        sed -e "/^# @ job_type = serial.*/d" \
            -e "/^# @ resources = threads_per_task(nb_openmp_processes_ll).*/d" \
            -e "/^# @ total_tasks = nb_mpi_processes_ll.*/d" \
        temp/job_zahir_exec_pid$$ > temp/job_zahir_exec_pid$$_bis ;

        mv temp/job_zahir_exec_pid$$_bis temp/job_zahir_exec_pid$$ ;
    elif [ $nb_openmp_processes -gt 0 -a $nb_mpi_processes -eq 0 ] ; then
        sed -e "/^# @ total_tasks = nb_mpi_processes_ll.*/d" \
        temp/job_zahir_exec_pid$$ > temp/job_zahir_exec_pid$$_bis ;

        mv temp/job_zahir_exec_pid$$_bis temp/job_zahir_exec_pid$$ ;
    elif [ $nb_openmp_processes -eq 0 -a $nb_mpi_processes -gt 0 ] ; then
        sed -e "s/^# @ job_type = serial.*/# @ job_type = parallel/g" \
            -e "/^# @ resources = threads_per_task(nb_openmp_processes_ll).*/d" \
        temp/job_zahir_exec_pid$$ > temp/job_zahir_exec_pid$$_bis ;

        mv temp/job_zahir_exec_pid$$_bis temp/job_zahir_exec_pid$$ ;
    else
        sed -e "s/^# @ job_type = serial.*/# @ job_type = parallel/g" \
        temp/job_zahir_exec_pid$$ > temp/job_zahir_exec_pid$$_bis ;

        mv temp/job_zahir_exec_pid$$_bis temp/job_zahir_exec_pid$$ ;
    fi

    sed -e "/^#!\/bin\/ksh.*/d" \
        -e "\%^. /etc/profile > /dev/null 2>&1.*%d" \
        -e "s/name_ll/$name_ll/g" \
        -e "s/time_cpu_ll/$time_cpu_ll/g" \
        -e "s/memory_data_ll/$memory_data_ll/g" \
        -e "s/memory_stack_ll/$memory_stack_ll/g" \
        -e "s/nb_openmp_processes_ll/$nb_openmp_processes/g" \
        -e "s/nb_mpi_processes_ll/$nb_mpi_processes/g" \
        temp/job_zahir_exec_pid$$ > temp/job_zahir_exec_pid$$_bis ;

    mv temp/job_zahir_exec_pid$$_bis temp/job_zahir_exec_pid$$ ;

fi

#_____________________________________________________
#
#       UPLOAD OF THE FILES ON ZAHIR COMPILATION ON ZAHIR
#_____________________________________________________
#

printf "\nUpload of the files on Zahir...\n"

# we create a directory on gaya to copy the executable after the compilation on Zahir.
dir_save_gaya_upload=$dir_save_gaya/ns3d_upload_${date_hour}_pid$$
ssh -x rbll200@gaya.idris.fr "mkdir -p $dir_save_gaya_upload"

# we compile in the $WORKDIR of Zahir
printf "\nTransfer of the files on Zahir for the compilation...\n"
printf "Directory of compilation: /workdir/rech/bll/rbll200/ns3d_compil_zahir_pid$$\n"
ssh -x rbll200@zahir.idris.fr "mkdir -p /workdir/rech/bll/rbll200/ns3d_compil_zahir_pid$$"
rcp -r     $dir_source/*                       rbll200@zahir.idris.fr:/workdir/rech/bll/rbll200/ns3d_compil_zahir_pid$$
rcp        temp/config.h_pid$$                 rbll200@zahir.idris.fr:/workdir/rech/bll/rbll200/ns3d_compil_zahir_pid$$/config.h
rcp        temp/data.in_pid$$                  rbll200@zahir.idris.fr:/workdir/rech/bll/rbll200/ns3d_compil_zahir_pid$$/data.in
rcp        temp/Makefile_pid$$                 rbll200@zahir.idris.fr:/workdir/rech/bll/rbll200/ns3d_compil_zahir_pid$$/Makefile


printf "\nCompilation on Zahir...\n"
ssh -x rbll200@zahir.idris.fr ". /etc/profile > /dev/null 2>&1 ; cd /workdir/rech/bll/rbll200/ns3d_compil_zahir_pid$$ ; make ; mfput ns3d  $dir_save_gaya_upload ; cd .. ;"


#_____________________________________
#
#       UPLOAD OF THE FILES ON GAYA
#_____________________________________
#

printf "\nUpload of the files on Gaya...\n"
# we upload all the necessary files (sources, data file, configuration file) in the working directory
rcp -r     $dir_source/*                       rbll200@gaya.idris.fr:$dir_save_gaya_upload
rcp        temp/config.h_pid$$                 rbll200@gaya.idris.fr:$dir_save_gaya_upload/config.h
rcp        temp/data.in_pid$$                  rbll200@gaya.idris.fr:$dir_save_gaya_upload/data.in
rcp        temp/Makefile_pid$$                 rbll200@gaya.idris.fr:$dir_save_gaya_upload/Makefile

# we upload the input data files if necessary
if [ -f "$velocity_init" ]; then
    rcp -v   $velocity_init         rbll200@gaya.idris.fr:$dir_save_gaya_upload/velocity.init
fi

if [ -f "$base2d_init" ]; then
    rcp -v   $base2d_init           rbll200@gaya.idris.fr:$dir_save_gaya_upload/base2D.init
fi

if [ $velocity_init_gaya ]; then
    printf "\n   Copy of the initial velocity from Gaya to Gaya...\n"
    ssh -x rbll200@gaya.idris.fr "cp -p $velocity_init_gaya $dir_save_gaya_upload/velocity.init"
fi



#_______________________________________________
#
#       UPLOAD OF THE EXECUTION SCRIPT ON ZAHIR
#_______________________________________________
#

printf "\nUpload of the execution script on Zahir...\n"

# we create a directory to save the execution script and we upload it
ssh -x rbll200@zahir.idris.fr "mkdir -p ns3d_job_zahir"
rcp temp/job_zahir_exec_pid$$ "rbll200@zahir.idris.fr:ns3d_job_zahir"

# we delete the useless configuration files
rm temp/config.h_pid$$ temp/data.in_pid$$ temp/job_zahir_exec_pid$$



#_____________________________________________________
#
#       EXECUTION / SUBMISSION OF THE JOB
#_____________________________________________________
#

# we are in batch mode: we submit the execution script
if [ "$mode" = "batch" ] ; then
    printf "\nSubmission of the job with LoadLeveler...\n"
    ssh -x rbll200@zahir.idris.fr "cd ns3d_job_zahir ; chmod +x job_zahir_exec_pid$$ ; llsubmit job_zahir_exec_pid$$"

    printf "\n********** END OF THE NS3D SCRIPT *********\n"

# we are in interactive mode: we do NOTHING! The execution script which was uploaded should be executed manually on Zahir.
elif [ "$mode" = "interactive" ] ; then
    ssh -x rbll200@zahir.idris.fr "cd ns3d_job_zahir ; chmod +x job_zahir_exec_pid$$"
    printf "\nIt is time to connect on Zahir and launch the execution script ns3d_job_zahir/job_zahir_exec_pid$$\n"

else
    printf "\nERROR : the execution mode (batch or interactive) is incorrect\n"
fi

exit
