#!/bin/sh

#
# This script makes easier using the NS3D code.
# It performs all the necessary steps:
#     - copy of the necessary files (input data, sources...) into the execution directory
#     - compilation
#     - execution
#     - save of the results in the save directory
#
# This basic version of the script is only valid for non parallel run.
#
# The argument "-s" (for instance "./job.sh -s") allows to display the code output on the screen.
# Without this argument, NS3D runs in silent mode.

#____________________________________________________________
#
#    PATHS/NAMES
#____________________________________________________________
#

# ------------------------------
# Directories/files used
# ------------------------------

# directory where the results are saved (a sub-directory is automatically created for each run)
dir_save="$HOME/NS3D_save"

# directory where the source files are
dir_source="../source"

# working directory where the code runs (a sub-directory is automatically created for each run)
dir_work="$HOME/NS3D_work"

# optional paths to the input data files (previous results data file to restart a run, initial velocity, base flow)
velocity_init=""
base2d_init=""

# name of the simulation, used to build the name the working/save folders
run_name="tanh"

# Makefile to use for the compilation (examples are provided in "source/Makefile_examples")
# if left blank, the file "Makefile" present in dir_source is used
makefile_file=""

#____________________________________________________________
#
#    COMPILATION PARAMETERS (config.h)
#____________________________________________________________
#

mkdir -p temp
cat > temp/config.h_pid$$ << EOF
! *******************************************
!    NS3D: compilation parameters
! *******************************************

! This file is used by the pre-processor during the compilation step.

! Avoid multiple inclusions of this file. Do not change!
#ifndef CONFIG_H
#define CONFIG_H

! Number of colocation points
#define DIMX 64 
#define DIMY 256
#define DIMZ 3

! Padding along each direction
#define PPADKX 0
#define PPADKY 0
#define PPADKZ 0

! The following time schemes are available:
!     - AB2: Adams-Bashforth of order 2
!     - RK2: Runge and Kutta of order 2
!     - RK3: Runge and Kutta of order 3
!     - RK4: Runge and Kutta of order 4
#define RK4

#endif
EOF



#____________________________________________________________
#
#    SIMULATION PARAMETERS (data.in)
#____________________________________________________________
#

cat > temp/data.in_pid$$ << EOF
********************************************
*    NS3D: simulation parameters           *
********************************************

This file is read at every simulation start.

*** discretization variables ***
lx___________________________   200.
ly___________________________   60.
lz___________________________   12.5664
dt___________________________   0.5
begin________________________   0.
itmax________________________   2000
de_aliasing__________________   T
    squared:1_or_elliptic:2__   1
    radius_truncation_x______   0.66
    radius_truncation_y______   0.66
    radius_truncation_z______   2.

*** physical variables ***
viscosity____________________   5.E-7
stratified___________________   F
    brunt_vaisala_frequency__   10.
    schmidt_number___________   1.
2omega_______________________   0.

*** Type of simulation ***
perturbative_________________   T
    linear___________________   T

*** Base state (only perturbation run) ***
base2D_type__________________   tanh

*** Initial velocity ***
velo_type____________________   null
white_noise__________________   1E-10

*** Output ***
output1_period_______________   400
output2_period_______________   1000
output3_period_______________   0

*** additionnal variables... ***

* Stuart vortices *
concentration_stuart_________   0.25

* File gaussian vortices *
file_nb_vortices_____________   2
Vortex 1
    position_x_______________   3.1415927
    position_y_______________   3.8165927
    circulation______________   2.
    core_radius______________   0.2
Vortex 2
    position_x_______________   3.1415927
    position_y_______________   2.4665927
    circulation______________   2.
    core_radius______________   0.2

* Random gaussian vortices *
rnd_nb_vortices______________   10
rnd_min_distance_____________   1.
rnd_mean_gamma_______________   6.28
rnd_std_gamma________________   0.
rnd_mean_radius______________   1.
rnd_std_radius_______________   0.

EOF

#____________________________________________________________
#
#     DO NOT MODIFY BELOW!
#____________________________________________________________
#

printf "\n********** START OF THE NS3D SCRIPT *********\n"


#_____________________________________________________
#
#    PRELIMINARY TASKS
#_____________________________________________________
#

# we read the run-time parameters and we configure the file config.h accordingly
NX=`grep "#define DIMX" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMX/ {print $3}'`
NY=`grep "#define DIMY" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMY/ {print $3}'`
NZ=`grep "#define DIMZ" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMZ/ {print $3}'`
LX=`awk '/^lx/ {print $2}' temp/data.in_pid$$`
LY=`awk '/^ly/ {print $2}' temp/data.in_pid$$`
LZ=`awk '/^lz/ {print $2}' temp/data.in_pid$$`
Dt=`awk '/^dt/ {print $2}' temp/data.in_pid$$`
nu=`awk '/^viscosity/ {print $2}' temp/data.in_pid$$`
stratifie=`awk '/^stratified/ {print $2}' temp/data.in_pid$$`
brunt=`awk '/brunt_vaisala_frequency/ {print $2}' temp/data.in_pid$$`

if [ "$stratifie" = "F" ]; then
    brunt=0
fi

# if the makefile_file variable is not defined, we search for a suitable Makefile
# the search is based on the host name
if [ -z "$makefile_file" ]; then
    printf "No Makefile specified...\n"

    if [ -n "`echo $HOSTNAME | grep -e raspoutine -e dracula -e joebar -e mandrake -e jill`" ]; then
        makefile_file=${dir_source}/Makefile_examples/Makefile_Intel64_fftw
    elif [ -n "`echo $HOSTNAME | grep -e tetsuo -e kelvin -e gailuron`" ]; then
        makefile_file=${dir_source}/Makefile_examples/Makefile_Intel32_fftw
    elif [ -n "`echo $HOSTNAME | grep -e rahan`" ]; then
        makefile_file=${dir_source}/Makefile_examples/Makefile_Mac_fftw
    elif [ -n "`echo $HOSTNAME | grep -e gibi -e melusine`" ]; then
        makefile_file=${dir_source}/Makefile_examples/Makefile_SGI_fftw
    else
    		makefile_file=${dir_source}/Makefile
    fi
fi

printf "Makefile: ${makefile_file}\n"

cp $makefile_file temp/Makefile_pid$$


#_____________________________________________________
#
#    COPY OF THE FILES IN THE WORKING DIRECTORY
#_____________________________________________________
#

printf "Copy of the files in the working directory...\n"

# we define the full name of the working directory and create it
date_heure=`date +%d-%m-%y_%Hh%M`
dir_work_full=${dir_work}/ns3d_work_${run_name}_${NX}x${NY}x${NZ}_L=${LX}x${LY}x${LZ}_Dt=${Dt}_nu=${nu}_N=${brunt}_${date_heure}_pid$$

mkdir -p $dir_work_full

# we copy all the necessary files (sources, data file, configuration file) in the working directory
cp -r  $dir_source/*                         $dir_work_full/
cp -f  temp/Makefile_pid$$                   $dir_work_full/Makefile
mv -f  temp/config.h_pid$$                   $dir_work_full/config.h
mv -f  temp/data.in_pid$$                    $dir_work_full/data.in

# we copy the input data files if necessary
if [ -f "$base2d_init" ]; then
    cp    $base2d_init        $dir_work_full/base2D.init
fi

if [ -f "$velocity_init" ]; then
    cp    $velocity_init      $dir_work_full/velocity.init
fi


#_____________________________________________________
#
#    COMPILATION
#_____________________________________________________
#

printf "Compilation...\n"

# we compile
cd $dir_work_full
make ns3d
chmod 744 *

#_____________________________________________________
#
#    LAUNCHING THE RUN
#_____________________________________________________
#

# we print everything in the file "listing.txt"
printf "\nStart of the simulation on $HOSTNAME: `date '+%A %d %B %Y a %T'`\n" | tee -a listing.txt
echo "*************************************************************************" >> listing.txt

echo " " >> listing.txt
echo " " >> listing.txt

echo "COMPILATION PARAMETERS" >> listing.txt
echo "***************************" >> listing.txt
cat config.h >> listing.txt
echo " " >> listing.txt
cat extended_config.h >> listing.txt

echo " " >> listing.txt
echo " " >> listing.txt

echo "SIMULATION PARAMETERS" >> listing.txt
echo "***************************" >> listing.txt
cat data.in >> listing.txt

echo " " >> listing.txt
echo " " >> listing.txt

echo " "
ulimit -s unlimited
if [ "$1" = "-s" ] ; then
    time nice -19 ./ns3d | tee -a listing.txt
else
    time nice -19 ./ns3d >> listing.txt
fi

echo " " >> listing.txt
echo " " >> listing.txt

printf "\nEnd of the simulation: `date '+%A %d %B %Y a %T'`\n" | tee -a listing.txt
echo "*************************************************************************" >> listing.txt

#_____________________________________________________
#
#     SAVING THE RESULTS
#_____________________________________________________
#

printf "\nSave of the results...\n"

# we create the saving directory
dir_save_full=$dir_save/ns3d_${run_name}_${NX}x${NY}x${NZ}_L=${LX}x${LY}x${LZ}_Dt=${Dt}_nu=${nu}_N=${brunt}_$date_heure

if [ -d "$dir_save_full" ] ; then
    dir_save_full=$dir_save_full"_"pid$$
fi

mkdir -p $dir_save_full

# we delete the executable, the object files and the module files. We save all the other files.
rm -f ns3d *.mod *.o base2D.init
mv * $dir_save_full

# we delete the working directory
cd ..
rm -rf $dir_work_full

is_empty=`ls -A $dir_work`
if [ -z "$is_empty" ]; then
    cd ..
    rm -rf $dir_work;
fi

printf "\n********** END OF THE NS3D SCRIPT *********\n"

exit
