!******************************************************************************
!
! NS3D: a Navier-Stokes solver for incompressible and stratified flow
!
! Organisation : LadHyX, Ecole Polytechnique, France
!
! Permission is granted to copy and distribute this file or modified
! versions of this file for no fee, provided the copyright notice and
! this permission notice are preserved on all copies.
!
!******************************************************************************

#include "config.h"
#include "extended_config.h"

!****************************************************************************
module gen_velocity
!****************************************************************************
!****h* ns3d/gen_velocity
!  NAME
!    base_state
!
!  DESCRIPTION
!    Generate velocity and vorticity fields.
!***

    use global_vars
    use fft
    use data_parser
    use subfunctions

    implicit none

    contains

! ***********************************************************
    subroutine gen_base2d(vbx,vby,vbz,wbx,wby,wbz,field_name,work)
! ***********************************************************
    !****f* gen_velocity/gen_base2d
    !  NAME
    !    gen_base2d
    !
    !  DESCRIPTION
    !    Generate a velocity and vorticity 2D base field
    !
    !  INPUT
    !    field_name: name of the field to generate
    !    work: working array
    !
    !  OUTPUT
    !    vbx,vby,vbz: flow velocity
    !    wbx,wby,wbz: flow vorticity
    !    work: the working array can be modified by gen_velo
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1), intent(out) :: vbx,vby,vbz,wbx,wby,wbz
        character(len=*), intent(in) :: field_name
        double precision, dimension(0:dx-1,0:dy-1,0:dz-1,*), intent(out) :: work

        if (field_name=='file') then
            call read_base2D(vbx,vby,vbz,wbx,wby,wbz)

        else if (field_name=='stuart') then
            call gen_velo_stuart(vbx,vby,vbz,wbx,wby,wbz)

        else if (field_name=='tanh') then
            call gen_velo_tanh(vbx,vby,vbz,wbx,wby,wbz)

        else if (field_name=='file_vortices') then
            call gen_velo_file_vortices(vbx,vby,vbz,wbx,wby,wbz,work(:,:,:,1),work(:,:,:,2),work(:,:,:,3))

        else if (field_name=='random_vortices') then
            call gen_velo_random_vortices(vbx,vby,vbz,wbx,wby,wbz,work(:,:,:,1),work(:,:,:,2),work(:,:,:,3))

        else if (field_name=='null') then
            vbx = 0.
            vby = 0.
            vbz = 0.
            wbx = 0.
            wby = 0.
            wbz = 0.

        end if

    end subroutine


! ***********************************************************
    subroutine gen_velo(vx,vy,vz,rho,field_name,work)
! ***********************************************************
    !****f* gen_velocity/gen_velo
    !  NAME
    !    gen_velo
    !
    !  DESCRIPTION
    !    Generate a velocity and vorticity 2D base field
    !
    !  INPUT
    !    field_name: name of the field to generate
    !    work: working array
    !
    !  OUTPUT
    !    vx,vy,vz,rho: flow velocity and density
    !    work: the working array is modified by gen_velo
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(out) :: vx,vy,vz,rho
        character(len=*), intent(in) :: field_name
        double precision, dimension(0:dx-1,0:dy-1,0:dz-1,*), intent(out) :: work

        if (field_name=='file') then
            call read_velo(vx,vy,vz,rho,begin)

        else if (field_name=='stuart') then
            call gen_velo_stuart(vx(:,:,0),vy(:,:,0),vz(:,:,0),work(:,:,0,1),work(:,:,0,2),work(:,:,0,3))
            rho(:,:,0) = 0.

            call z_invariant(vx,vy,vz,rho)

        else if (field_name=='tanh') then
            call gen_velo_tanh(vx(:,:,0),vy(:,:,0),vz(:,:,0),work(:,:,0,1),work(:,:,0,2),work(:,:,0,3))
            rho(:,:,0) = 0.
            call z_invariant(vx,vy,vz,rho)

        else if (field_name=='file_vortices') then
            call gen_velo_file_vortices(vx(:,:,0),vy(:,:,0),vz(:,:,0),work(:,:,0,1),work(:,:,0,2),work(:,:,0,3), &
                                        work(:,:,:,4),work(:,:,:,5),work(:,:,:,6))
            call z_invariant(vx,vy,vz,rho)

        else if (field_name=='random_vortices') then
            call gen_velo_random_vortices(vx(:,:,0),vy(:,:,0),vz(:,:,0),work(:,:,0,1),work(:,:,0,2),work(:,:,0,3), &
                                          work(:,:,:,4), work(:,:,:,5),work(:,:,:,6))
            rho(:,:,0) = 0.
            call z_invariant(vx,vy,vz,rho)

        else if (field_name=='null') then
            vx = 0.
            vy = 0.
            vz = 0.
            rho = 0.

        end if

    end subroutine

! ********************************************************************************
    subroutine z_invariant(vx,vy,vz,rho)
! ********************************************************************************
    !****f* gen_velocity/z_invariant
    !  NAME
    !    z_invariant
    !
    !  DESCRIPTION
    !    Generate a velocity invariant in the z-direction
    !
    !  INPUT
    !    vx,vy,vz,rho: the level iz=0 will be used as a reference
    !
    !  OUTPUT
    !    vx,vy,vz,rho; velocity invariant in the z-direction
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(inout) :: vx, vy, vz,rho
        integer :: iz

      do iz=1,nz-1
            vx(:,:,iz) = vx(:,:,0)
            vy(:,:,iz) = vy(:,:,0)
            vz(:,:,iz) = vz(:,:,0)
            rho(:,:,iz) = rho(:,:,0)
      end do

    end subroutine


! ********************************************************************************
    subroutine read_base2D(vbx,vby,vbz,wbx,wby,wbz)
! ********************************************************************************
    !****f* gen_velocity/read_base2D
    !  NAME
    !    read_base2D
    !
    !  DESCRIPTION
    !    Read a base flow in the binary file "base2D.init".
    !
    !  OUTPUT
    !    vbx,vby,vbz: base flow velocity
    !    wbx,wby,wbz: base flow vorticity
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1), intent(out) :: vbx, vby, vbz, wbx, wby, wbz
        integer :: flag, nxread, nyread
        double precision :: lxread, lyread

        open (unit=88,file='base2D.init',form='unformatted',action='read')
        read(88) flag, nxread, nyread, lxread, lyread

        ! we check the file format: we check only the dimensions that must not changed between two runs
        if (flag/=1 .or. nxread/=gnx .or. nyread/=gny .or. abs(lxread-lx)>epsilo .or. abs(lyread-ly)>epsilo) then
            write(*,*) "ERROR: the 2D-base-flow file has bad format. Failure of the simulation."
            stop
        end if

        read(88) vbx(0:nx-1,0:ny-1), vby(0:nx-1,0:ny-1), vbz(0:nx-1,0:ny-1)
        read(88) wbx(0:nx-1,0:ny-1), wby(0:nx-1,0:ny-1), wbz(0:nx-1,0:ny-1)
        close(88)


    end subroutine

#ifdef MPI
! ***********************************************************
    subroutine read_velo(pvx,pvy,pvz,prho,instant)
! ***********************************************************
    !****f* gen_velocity/read_velo
    !  NAME
    !    read_velo
    !
    !  DESCRIPTION
    !    read the results from a previous run in a binary file "velocity.init"
    !    This is the MPI version of the previous sequential subroutine. It uses the MPI-IO (MPI input-output library).
    !
    !  OUTPUT
    !    pvx,pvy,pvz,prho: initial velocity and density fields of the restarted run in physical space
    !    instant: time of the simulation
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1),intent(out) :: pvx,pvy,pvz,prho
        double precision, intent(out) :: instant
        integer :: flag, gnxread, gnyread, gnzread
        double precision :: lxread, lyread, lzread
        double precision :: trash_double
        integer :: trash_int
        logical :: trash_log
        integer :: descriptor
        integer, dimension(MPI_STATUS_SIZE) :: mpi_status

        call MPI_file_open(MPI_COMM_WORLD,'velocity.init',MPI_MODE_RDONLY,MPI_INFO_NULL,descriptor,code)

        ! only the process rank 0 read the header and check the format of the file
        if (rank==0) then
            call MPI_file_read_shared(descriptor,trash_int,        1, MPI_INTEGER,mpi_status,code)
            call MPI_file_read_shared(descriptor,flag,             1, MPI_INTEGER,mpi_status,code)
            call MPI_file_read_shared(descriptor,gnxread,          1, MPI_INTEGER,mpi_status,code)
            call MPI_file_read_shared(descriptor,gnyread,          1, MPI_INTEGER,mpi_status,code)
            call MPI_file_read_shared(descriptor,gnzread,          1, MPI_INTEGER,mpi_status,code)
            call MPI_file_read_shared(descriptor,lxread,           1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_read_shared(descriptor,lyread,           1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_read_shared(descriptor,lzread,           1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_read_shared(descriptor,trash_double,     1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_read_shared(descriptor,trash_log,        1, MPI_LOGICAL          ,mpi_status,code)
            call MPI_file_read_shared(descriptor,trash_int,        1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_read_shared(descriptor,trash_double,     1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_read_shared(descriptor,trash_double,     1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_read_shared(descriptor,trash_double,     1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_read_shared(descriptor,trash_double,     1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_read_shared(descriptor,trash_log,        1, MPI_LOGICAL          ,mpi_status,code)
            call MPI_file_read_shared(descriptor,trash_double,     1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_read_shared(descriptor,trash_double,     1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_read_shared(descriptor,trash_double,     1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_read_shared(descriptor,trash_log,        1, MPI_LOGICAL          ,mpi_status,code)
            call MPI_file_read_shared(descriptor,trash_log,        1, MPI_LOGICAL          ,mpi_status,code)
            call MPI_file_read_shared(descriptor,trash_int,        1, MPI_INTEGER          ,mpi_status,code)

            call MPI_file_read_shared(descriptor,trash_int,        1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_read_shared(descriptor,instant,          1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_read_shared(descriptor,trash_int,        1, MPI_INTEGER          ,mpi_status,code)

           if (flag /= 1 .or. gnxread /= gnx .or. gnyread /= gny .or. gnzread /= gnz .or. &
               abs(lxread-lx)>epsilo .or. abs(lyread-ly)>epsilo) then
               write(*,*) "ERROR: the restarting-run file has bad format. Failure of the simulation."
               call mpi_abort(MPI_COMM_WORLD, 0)
           end if
        end if

        ! each process reads only a slide of the velocity, density and vorticity fields
        ! the process rank 0 reads the integer value that corresponds to the size of the record
        if (rank==0) then
            call MPI_file_read_shared(descriptor,trash_int,1, MPI_INTEGER,mpi_status,code)
        end if
        call MPI_file_read_ordered(descriptor,pvx(0:nx-1,0:ny-1,0:nz-1),nx*ny*nz,MPI_DOUBLE_PRECISION,mpi_status,code)
        if (rank==0) then
            call MPI_file_read_shared(descriptor,trash_int,1, MPI_INTEGER,mpi_status,code)
        end if

        if (rank==0) then
            call MPI_file_read_shared(descriptor,trash_int,1, MPI_INTEGER,mpi_status,code)
        end if
        call MPI_file_read_ordered(descriptor,pvy(0:nx-1,0:ny-1,0:nz-1),nx*ny*nz,MPI_DOUBLE_PRECISION,mpi_status,code)
        if (rank==0) then
            call MPI_file_read_shared(descriptor,trash_int,1, MPI_INTEGER,mpi_status,code)
        end if

        if (rank==0) then
            call MPI_file_read_shared(descriptor,trash_int,1, MPI_INTEGER,mpi_status,code)
        end if
        call MPI_file_read_ordered(descriptor,pvz(0:nx-1,0:ny-1,0:nz-1),nx*ny*nz,MPI_DOUBLE_PRECISION,mpi_status,code)
        if (rank==0) then
            call MPI_file_read_shared(descriptor,trash_int,1, MPI_INTEGER,mpi_status,code)
        end if

        if (rank==0) then
            call MPI_file_read_shared(descriptor,trash_int,1, MPI_INTEGER,mpi_status,code)
        end if
        call MPI_file_read_ordered(descriptor,prho(0:nx-1,0:ny-1,0:nz-1),nx*ny*nz,MPI_DOUBLE_PRECISION,mpi_status,code)
        if (rank==0) then
            call MPI_file_read_shared(descriptor,trash_int,1, MPI_INTEGER,mpi_status,code)
        end if

        call MPI_file_close (descriptor,code)

    end subroutine


#else
! ***********************************************************
    subroutine read_velo(pvx,pvy,pvz,prho,instant)
! ***********************************************************
    !****f* gen_velocity/read_velo
    !  NAME
    !    read_velo
    !
    !  DESCRIPTION
    !    read the results from a previous run in a binary file "velocity.init"
    !    in order to restart the simulation.
    !
    !  OUTPUT
    !    pvx,pvy,pvz,prho: initial velocity and density fields of the restarted run in physical space
    !    instant: time of the simulation
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1),intent(out) :: pvx,pvy,pvz,prho
        double precision, intent(out) :: instant
        integer :: flag, nxread, nyread, nzread
        double precision :: lxread, lyread, lzread
        double precision :: trash_double
        integer :: trash_int
        logical :: trash_log

        open (unit=8,file='velocity.init',form='unformatted')

        ! we read the meta-data
        read(8) flag,nxread,nyread,nzread,lxread,lyread,lzread,trash_double,trash_log,trash_int,trash_double,trash_double, &
                trash_double,trash_double,trash_log,trash_double,trash_double,trash_double,trash_log,trash_log

        read(8) instant

        ! we check the file format: we check only the dimensions that must not be changed between two runs
        if (flag /= 1 .or. nxread /= gnx .or. nyread /= gny .or. nzread /= gnz .or. &
            abs(lxread-lx)>epsilo .or. abs(lyread-ly)>epsilo) then
            write(*,*) "ERROR: the restarting-run file has bad format. Failure of the simulation."
            stop
        end if

        call read_array(pvx,8)
        call read_array(pvy,8)
        call read_array(pvz,8)
        call read_array(prho,8)

        close(8)

    end subroutine


! ***********************************************************
    subroutine read_array(pa,id_file)
! ***********************************************************
    !****f* gen_velocity/read_array
    !  NAME
    !    read_array
    !
    !  DESCRIPTION
    !    Allow to read a subsection of an array from a file without using too much stack memory which
    !    can be an issue on some systems with limited stack memory.
    !    Indeed, surprisingly, on some systems (NEC-SX8) the following instructions are responsible for the
    !    use of an amount of stack memory corresponding to the **sum** of all the recorded arrays:
    !        read(1) aax(0:nx-1,0:ny-1,0:nz-1)
    !        read(1) aay(0:nx-1,0:ny-1,0:nz-1)
    !        (...)
    !
    !    This subroutine makes use of a single storage array that is static and so defined in data memory.
    !
    !  INPUTS
    !    pa: field in physical space where to store the file record
    !    id_file: identifier of the file where the field should be read
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1),intent(out) :: pa
        integer, intent(in) :: id_file
        integer :: ix, iy, iz

        read(id_file) wk_io

        do iz=0,nz-1
            do iy=0,ny-1
                do ix=0,nx-1
                    pa(ix,iy,iz) = wk_io(ix,iy,iz)
                end do
            end do
        end do

    end subroutine
#endif

! ***********************************************************
    subroutine gen_velo_stuart(vx,vy,vz,wx,wy,wz)
! ***********************************************************
    !****f* gen_velocity/gen_velo_stuart
    !  NAME
    !    gen_velo_stuart
    !
    !  DESCRIPTION
    !    Generate a velocity and vorticity field corresponding to Stuart vortices
    !
    !  INPUT
    !    rho_stuart: Stuart concentration parameter
    !
    !  OUTPUT
    !    vx,vy,vz: flow velocity
    !    wx,wy,wz: flow vorticity
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1), intent(out) :: vx,vy,vz,wx,wy,wz
        double precision :: rho_stuart
        integer :: ix,iy

        vx = 0.
        vy = 0.
        vz = 0.
        wx = 0.
        wy = 0.
        wz = 0.

        ! we read the concentration parameter in the data file 'data.in'
        call parse('data.in','concentration_stuart',rho_stuart,0.25d0)

        do iy=0,ny-1
            do ix=0,nx-1
                vx(ix,iy) = sinh(2.D0*(yy(iy)-ly/2.D0))/(cosh(2.D0* (yy(iy)-ly/2.D0)) - rho_stuart*cos(2.D0*xx(ix)))
                vy(ix,iy) = - rho_stuart * sin( 2.D0*xx(ix) )/( cosh( 2.D0*( yy(iy)-ly/2.D0 ) ) - rho_stuart*cos(2.D0*xx(ix)))
                wz(ix,iy) = 2.D0*(rho_stuart**2-1.D0)/(cosh(2.D0*(yy(iy)-ly/2.D0)) - rho_stuart*cos(2.D0*xx(ix)))**2
            end do
        end do

    end subroutine


! ***********************************************************
    subroutine gen_velo_tanh(vx,vy,vz,wx,wy,wz)
! ***********************************************************
    !****f* gen_velocity/gen_velo_tanh
    !  NAME
    !    gen_velo_tanh
    !
    !  DESCRIPTION
    !    Generate a velocity and vorticity field corresponding to the hyperbolic-tangent mixing layer
    !
    !  INPUT
    !    nx,ny,nz: physical dimensions
    !    dx,dy,dz: storage dimensions
    !
    !  OUTPUT
    !    vx,vy,vz: flow velocity
    !    wx,wy,wz: flow vorticity
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1), intent(out) :: vx,vy,vz,wx,wy,wz
        integer :: ix

        vx = 0.
        vy = 0.
        vz = 0.
        wx = 0.
        wy = 0.
        wz = 0.

        do ix=0,nx-1
            vx(ix,:) = tanh(yy-ly/2.D0)
            wz(ix,:) = -(1.D0-tanh(yy-ly/2.D0)**2)
        end do

    end subroutine


! ***********************************************************
    subroutine gen_velo_vortices(vx,vy,vz,wx,wy,wz,nb_vortices,xt,yt,gamma,at,work_vx,work_vy,work_wz)
! ***********************************************************
    !****f* gen_velocity/gen_velo_vortices
    !  NAME
    !    gen_velo_vortices
    !
    !  DESCRIPTION
    !    Generate a velocity and vorticity field corresponding to the superposition of vortices whose characteristics
    !    are given in the file 'data.in'
    !
    !  INPUT
    !    work: working field
    !
    !  OUTPUT
    !    vx,vy,vz: flow velocity
    !    wx,wy,wz: flow vorticity
    !    work: the working array is modified by gen_velo_vortices
    !***
    ! Additionnal variables !

    ! gaussian vortices

        implicit none

        double precision, dimension(0:dx-1,0:dy-1), intent(out) :: vx,vy,vz,wx,wy,wz
        integer, intent(in) :: nb_vortices
        double precision, dimension(nb_vortices), intent(in) :: xt, yt, gamma, at
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: work_vx,work_vy,work_wz
        double precision :: xt_current, yt_current
        double precision :: pi, r2, k2
        integer :: ix_box, iy_box, i_vortex
        integer :: ix,iy
        integer :: ik1,ik2,ik3

        pi = acos(-1.D0)

        vx = 0.
        vy = 0.
        vz = 0.
        wx = 0.
        wy = 0.
        wz = 0.

        ! vertical vorticity of the flow.
        ! To ensure the spatial periodicity of wz, we consider artifical adjacent boxes identical to the main central box.
        do ix_box = -2,2
            do iy_box = -2,2
                do i_vortex=1,nb_vortices
                    do iy=0,ny-1
                        do ix=0,nx-1
                            xt_current = xt(i_vortex) + ix_box*lx
                            yt_current = yt(i_vortex) + iy_box*ly

                            r2 = (xx(ix)-xt_current)**2 + (yy(iy)-yt_current)**2

                            wz(ix,iy)=wz(ix,iy)+gamma(i_vortex)/(pi*at(i_vortex)**2)*exp(-r2/at(i_vortex)**2)
                        end do
                    end do
                end do
            end do
        end do

#ifdef MPI
        call MPI_bcast(wz,dx*dy, MPI_DOUBLE_PRECISION ,0,MPI_COMM_WORLD ,code)
#endif

        ! inversion of the 2D laplacian to find the stream-function 2D*1D
        call copy_p2D_to_p3D(wz,work_wz)
        call fwd_fft(work_wz,fft_work)

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    k2=max(kx2(IKX)+ky2(IKY),epsilo)
                    work_wz(:,ik1,ik2,ik3)= work_wz(:,ik1,ik2,ik3)/k2
                end do
            end do
        end do

        ! velocity corresponding to the stream-function 2D*1D
        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    work_vx(1,ik1,ik2,ik3) = -ky(IKY)*work_wz(2,ik1,ik2,ik3)
                    work_vx(2,ik1,ik2,ik3) =  ky(IKY)*work_wz(1,ik1,ik2,ik3)

                    work_vy(1,ik1,ik2,ik3) =  kx(IKX)*work_wz(2,ik1,ik2,ik3)
                    work_vy(2,ik1,ik2,ik3) = -kx(IKX)*work_wz(1,ik1,ik2,ik3)
                end do
            end do
        end do

        !  we switch into physical space
        call bck_fft(work_vx,fft_work)
        call bck_fft(work_vy,fft_work)

        call copy_p3D_to_p2D(work_vx,vx)
        call copy_p3D_to_p2D(work_vy,vy)

#ifdef MPI
        call MPI_bcast(vx,dx*dy, MPI_DOUBLE_PRECISION ,0,MPI_COMM_WORLD ,code)
        call MPI_bcast(vy,dx*dy, MPI_DOUBLE_PRECISION ,0,MPI_COMM_WORLD ,code)
#endif

    end subroutine


! ***************************************************************
    subroutine copy_p2D_to_p3D(p2D,p3D)
! ***************************************************************
    !****f* gen_velocity/copy_p2D_to_p3D
    !  NAME
    !    copy_p2D_to_p3D
    !
    !  DESCRIPTION
    !    Copy a physical 2D field p2D(x,y) into a z-invariant physical 3D field p3D(x,y,z)
    !
    !  INPUTS
    !    p2D: the physical 2D field p2D(x,y)
    !
    !  OUTPUT
    !    p3D: the physical 3D field p3D(x,y,z) which is invariant in z
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1), intent(in) :: p2D
        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(out) :: p3D
        integer :: iz

        do iz=0,nz-1
            p3D(:,:,iz) = p2D(:,:)
        end do

    end subroutine


! ***************************************************************
    subroutine copy_p3D_to_p2D(p3D,p2D)
! ***************************************************************
    !****f* gen_velocity/copy_p3D_to_p2D
    !  NAME
    !    copy_p3D_to_p2D
    !
    !  DESCRIPTION
    !    Copy a z-invariant physical 3D field p3D(x,y,z) into a physical 2D field p2D(x,y)
    !
    !  INPUTS
    !    p3D: the physical 3D field p3D(z,y,x) which is invariant in z
    !
    !  OUTPUT
    !    p2D: the physical 2D field p2D(y,x)
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(in) :: p3D
        double precision, dimension(0:dx-1,0:dy-1), intent(out) :: p2D

        p2D(:,:) = p3D(:,:,0)

    end subroutine

! ***********************************************************
    subroutine gen_velo_file_vortices(vx,vy,vz,wx,wy,wz,work_vx,work_vy,work_wz)
! ***********************************************************
    !****f* gen_velocity/gen_velo_file_vortices
    !  NAME
    !    gen_velo_file_vortices
    !
    !  DESCRIPTION
    !    Generate a velocity and vorticity field corresponding to the superposition of vortices whose characteristics
    !    are randomly determined
    !
    !  INPUT
    !    work: working field
    !
    !  OUTPUT
    !    vx,vy,vz: flow velocity
    !    wx,wy,wz: flow vorticity
    !    work_vx,work_vy,work_wz: the working arrays are modified by gen_velo_random_vortices
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1), intent(out) :: vx,vy,vz,wx,wy,wz
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: work_vx,work_vy,work_wz
        integer :: nb_vortices
        double precision, dimension(:), allocatable :: xt,yt,gamma,at

        ! we read in the data file 'data.in' the different characteristics of the random vortices
        call parse('data.in','file_nb_vortices',nb_vortices)

        ! we allocate the memory to store the array characteristics
        allocate(xt(nb_vortices), yt(nb_vortices), gamma(nb_vortices), at(nb_vortices))

        call parse('data.in','position_x',xt)
        call parse('data.in','position_y',yt)
        call parse('data.in','circulation',gamma)
        call parse('data.in','core_radius',at)

        ! we generate the corresponding vorticity and vorticity field
        call gen_velo_vortices(vx,vy,vz,wx,wy,wz,nb_vortices,xt,yt,gamma,at,work_vx,work_vy,work_wz)

    ! we de-allocate the memory used to store the array characteristics
        deallocate(xt,yt,gamma,at)

    end subroutine

! ***********************************************************
    subroutine gen_velo_random_vortices(vx,vy,vz,wx,wy,wz,work_vx,work_vy,work_wz)
! ***********************************************************
    !****f* gen_velocity/gen_velo_random_vortices
    !  NAME
    !    gen_velo_random_vortices
    !
    !  DESCRIPTION
    !    Generate a velocity and vorticity field corresponding to the superposition of vortices whose characteristics
    !    are randomly determined.
    !    We ensure a 0 mean vorticity by generating paris of vortices of opposed circulation and identical radius.
    !
    !  INPUT
    !    work: working field
    !
    !  OUTPUT
    !    vx,vy,vz: flow velocity
    !    wx,wy,wz: flow vorticity
    !    work_vx,work_vy,work_wz: the working arrays are modified by gen_velo_random_vortices
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1), intent(out) :: vx,vy,vz,wx,wy,wz
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: work_vx,work_vy,work_wz
        integer :: nb_vortices
        double precision :: min_distance, mean_gamma, std_gamma, mean_a, std_a
        double precision, dimension(:), allocatable :: xt,yt,gamma,at
        double precision :: current_x, current_y
        integer :: i, i_pair, current_nb_vortices
        logical :: test

        ! we read in the data file 'data.in' the different characteristics of the random vortices
        call parse('data.in','rnd_nb_vortices',nb_vortices)
        call parse('data.in','rnd_min_distance',min_distance)
        call parse('data.in','rnd_mean_gamma',mean_gamma)
        call parse('data.in','rnd_std_gamma',std_gamma)
        call parse('data.in','rnd_mean_radius',mean_a)
        call parse('data.in','rnd_std_radius',std_a)

        ! nb_vortices must be even in order to generate contra-rotating pairs of vortices and ensure a 0 mean vorticity.
        nb_vortices = 2*(nb_vortices/2)

        ! we allocate the memory to store the array characteristics
        allocate(xt(nb_vortices), yt(nb_vortices), gamma(nb_vortices), at(nb_vortices))

        ! we generate the characteristics of the random vortices
        current_nb_vortices = 0
        i_pair = 1

        do while (current_nb_vortices < nb_vortices)

            call random_number(current_x)
            current_x = lx*current_x

            call random_number(current_y)
            current_y = ly*current_y

            ! we check if the position of the current vortex is not too close from previous ones
            test = .true.
            do i=1,current_nb_vortices
                if ( (current_x-xt(i))**2+(current_y-yt(i))**2 < min_distance**2 ) then
                    test = .false.
                end if
            end do

            if (test) then
                current_nb_vortices = current_nb_vortices + 1

                xt(current_nb_vortices) = current_x
                yt(current_nb_vortices) = current_y

                ! first vortex of a pair: we randomly choose its circulation and radius
                if (i_pair==1) then
                    gamma(current_nb_vortices) = gaussian(mean_gamma,std_gamma)
                    at(current_nb_vortices) = gaussian(mean_a, std_a)
                    i_pair = 2

                ! second vortex of a pair: we take the opposed circulation and the same radius
                else
                    gamma(current_nb_vortices) = -gamma(current_nb_vortices-1)
                    at(current_nb_vortices) = at(current_nb_vortices-1)
                    i_pair = 1
                end if
            end if
        end do

        ! we generate the corresponding vorticity and vorticity field
        call gen_velo_vortices(vx,vy,vz,wx,wy,wz,nb_vortices,xt,yt,gamma,at,work_vx,work_vy,work_wz)

        ! we de-allocate the memory used to store the array characteristics
        deallocate(xt,yt,gamma,at)

    end subroutine


! *************************************************************************
    subroutine add_white_noise(pax,pay,paz,amplitude)
! *********************************************************
    !****f* gen_velocity/add_white_noise
    !  NAME
    !    add_white_noise
    !
    !  DESCRIPTION
    !    Add white noise to a field
    !
    !  INPUTS
    !    pax,pay,paz: field without noise in physical space
    !    amplitude: amplitude of the white noise
    !
    !  OUTPUT
    !    pax,pay,paz: field with white noise
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(inout) :: pax, pay, paz
        double precision, intent(in) :: amplitude
        double precision :: rnd_number
        double precision :: sumx_init, sumy_init, sumz_init, sumx_after, sumy_after, sumz_after
        integer :: iz, iy, ix

#ifdef MPI
        integer :: size_seed
        integer, dimension(:), allocatable :: iseed

        ! in MPI, we initialize the random seed differently for each process
        call random_seed(size=size_seed)
        allocate(iseed(size_seed))
        call random_seed(get=iseed)
        iseed = (modulo(iseed,nb_proc) + rank) * (iseed / (nb_proc+rank) )
        call random_seed(put=iseed)
        deallocate(iseed)
#endif

        ! we determine the sum of the fields
        sumx_init = sum(pax(0:nx-1,0:ny-1,0:nz-1))
        sumy_init = sum(pay(0:nx-1,0:ny-1,0:nz-1))
        sumz_init = sum(paz(0:nx-1,0:ny-1,0:nz-1))

#ifdef MPI
        call MPI_allreduce(MPI_IN_PLACE,sumx_init,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,sumy_init,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,sumz_init,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
#endif

        ! we add the white noise in physical space
        do iz=0,nz-1
            do iy=0,ny-1
                do ix=0,nx-1
                    call random_number(rnd_number)
                    pax(ix,iy,iz)=pax(ix,iy,iz)+amplitude*rnd_number

                    call random_number(rnd_number)
                    pay(ix,iy,iz)=pay(ix,iy,iz)+amplitude*rnd_number

                    call random_number(rnd_number)
                    paz(ix,iy,iz)=paz(ix,iy,iz)+amplitude*rnd_number
                end do
            end do
        end do

        ! in order to ensure a zero-mean noise we use the difference of means before and after the addition of noise
        sumx_after = sum(pax(0:nx-1,0:ny-1,0:nz-1))
        sumy_after = sum(pay(0:nx-1,0:ny-1,0:nz-1))
        sumz_after = sum(paz(0:nx-1,0:ny-1,0:nz-1))

#ifdef MPI
        call MPI_allreduce(MPI_IN_PLACE,sumx_after,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,sumy_after,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,sumz_after,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
#endif

        pax = pax + (sumx_init-sumx_after)/(gnx*gny*gnz)
        pay = pay + (sumy_init-sumy_after)/(gnx*gny*gnz)
        paz = paz + (sumz_init-sumz_after)/(gnx*gny*gnz)

    end subroutine

end module
