!******************************************************************************
!
! NS3D: a Navier-Stokes solver for incompressible and stratified flow
!
! Organisation : LadHyX, Ecole Polytechnique, France
!
! Permission is granted to copy and distribute this file or modified
! versions of this file for no fee, provided the copyright notice and
! this permission notice are preserved on all copies.
!
!******************************************************************************

#include "config.h"
#include "extended_config.h"

!****************************************************************************
module input
!****************************************************************************
!****h* ns3d/input
!  NAME
!    input
!
!  DESCRIPTION
!    Perform all the basic input operations on the parameter file (data.in)
!***

    use data_parser
    use global_vars

    implicit none

    contains


!***************************************************************************
    subroutine read_param
!***************************************************************************
    !****f* input/read_param
    !  NAME
    !    read_param
    !
    !  DESCRIPTION
    !    Read all the simulation parameters from the "data.in" file
    !***

        ! simulation sixe box
        call parse('data.in','lx',lx)
        call parse('data.in','ly',ly)
        call parse('data.in','lz',lz)

        ! time step and initial time
        call parse('data.in','dt',dt)
        call parse('data.in','begin',begin)

        ! number of time steps
        call parse('data.in','itmax',itmax)

        ! de_aliasing required for desaliasing
        call parse('data.in','de_aliasing',de_aliasing)
        call parse('data.in','squared:1_or_elliptic:2',type_trunc)
        call parse('data.in','radius_truncation_x',rtrunc_x)
        call parse('data.in','radius_truncation_y',rtrunc_y)
        call parse('data.in','radius_truncation_z',rtrunc_z)

        ! physical parameters
        call parse('data.in','viscosity',nu)
        call parse('data.in','stratified',stratification)
        call parse('data.in','brunt_vaisala_frequency',xns)
        call parse('data.in','schmidt_number',schm)
        call parse('data.in','2omega',omega2)
        xnsqr=xns**2

        ! type of simulation: in pertubation or not, linear or not
        call parse('data.in','perturbative',perturbative)
        call parse('data.in','linear',linear)

        ! input of initial velocity: reading data in a file or generating the flow
        ! white noise and its amplitude
        call parse('data.in','base2D_type',base2D_type)
        call parse('data.in','velo_type',velo_type)
        call parse('data.in','white_noise',white_noise)

        ! outputs
        call parse('data.in','output1_period',output1_period)
        call parse('data.in','output2_period',output2_period)
        call parse('data.in','output3_period',output3_period)

    end subroutine


!***************************************************************************
    subroutine print_param
!***************************************************************************
    !****f* input/print_param
    !  NAME
    !    print_param
    !
    !  DESCRIPTION
    !    Print all the simulation parameters
    !***

        implicit none

        character(len=100) :: aff_reel, aff_entier, aff_logical, aff_char

        ! printing formats
        aff_reel = '(A29,TR3,G12.5)'
        aff_entier = '(A29,TR4,I7)'
        aff_logical = '(A29,TR4,L1)'
        aff_char = '(A29,TR4,A40)'

        write(*,*) '----------------------------'
        write(*,*) 'DISCRETIZATION VARIABLES'
        write(*,*) '----------------------------'
        write(*,aff_reel) 'lx..........................',lx
        write(*,aff_reel) 'ly..........................',ly
        write(*,aff_reel) 'lz..........................',lz
        write(*,aff_reel) 'dt..........................',dt
        write(*,aff_reel) 'begin.......................',begin
        write(*,aff_entier) 'itmax.......................',itmax
        write(*,aff_logical) 'de_aliasing.................',de_aliasing
        write(*,aff_entier) '    squared:1_/_elliptic:2....',type_trunc
        write(*,aff_reel) '    radius_truncation_x......',rtrunc_x
        write(*,aff_reel) '    radius_truncation_y.......',rtrunc_y
        write(*,aff_reel) '    radius_truncation_z........',rtrunc_z

        write(*,*)
        write(*,*) '----------------------------'
        write(*,*) 'PHYSICAL VARIABLES'
        write(*,*) '----------------------------'
        write(*,aff_reel) 'viscosity...................',nu
        write(*,aff_logical) 'stratification..............',stratification
        write(*,aff_reel) '    brunt_vaisala_frequency..',xns
        write(*,aff_reel) '    schmidt_number...........',schm
        write(*,aff_reel) 'omega2......................',omega2

        write(*,*)
        write(*,*) '----------------------------'
        write(*,*) 'TYPE OF RUN'
        write(*,*) '----------------------------'
        write(*,aff_logical) 'perturbative................',perturbative
        write(*,aff_logical) '    linear..................',linear

        write(*,*)
        write(*,*) '----------------------------'
        write(*,*) 'BASE STATE'
        write(*,*) '----------------------------'
        write(*,aff_char) 'base2D_type.................',base2D_type

        write(*,*)
        write(*,*) '----------------------------'
        write(*,*) 'INITIAL VELOCITY'
        write(*,*) '----------------------------'
        write(*,aff_char) 'velo_type...................',velo_type
        write(*,aff_reel) 'white_noise.................',white_noise

        write(*,*)
        write(*,*) '----------------------------'
        write(*,*) 'OUTPUT'
        write(*,*) '----------------------------'
        write(*,aff_entier) 'output1_period..............',output1_period
        write(*,aff_entier) 'output2_period..............',output2_period
        write(*,aff_entier) 'output3_period..............',output3_period

        write(*,*)
        write(*,*)

    end subroutine

end module
