!******************************************************************************
!
! NS3D: a Navier-Stokes solver for incompressible and stratified flow
!
! Organisation : LadHyX, Ecole Polytechnique, France
!
! Permission is granted to copy and distribute this file or modified
! versions of this file for no fee, provided the copyright notice and
! this permission notice are preserved on all copies.
!
!******************************************************************************

#include "config.h"
#include "extended_config.h"

!****************************************************************************
module output
!****************************************************************************
!****h* ns3d/output
!  NAME
!    output
!
!  DESCRIPTION
!    Perform all the ouput operations
!***

    use global_vars
    use fft
    use subfunctions

    implicit none

    contains

! *************************************************************************
    subroutine output_base(vbx,vby,vbz,wbx,wby,wbz)
! *************************************************************************
    !****f* output/output_base
    !  NAME
    !    output_base
    !
    !  DESCRIPTION
    !    Record in a binary file the velocity and vorticity of the base state in physical space.
    !
    !  INPUTS
    !    vbx,vby,vbz,wbx,wby,wbz: velocity and vorticity of the base fields in physical space
    !***
        double precision, dimension(0:dx-1,0:dy-1), intent(in) :: vbx,vby,vbz,wbx,wby,wbz

        ! we first write meta-data in order to help the post-treatment of the data file and avoid errors
        ! Fortran adds before and after every recording an integer that corresponds to the size of the recording
        ! expressed in bytes. We must also take into account these extra integers while post-treating the file.

        ! the exact format file is (i and j are integers corresponding to the preceding or following recording):
        ! i,1,gnx,gny,lx,ly,j,i,vbx,vby,vbz,j,i,wbx,wby,wbz,j

        open (77,file='base2D.init',form='unformatted')
        write(77) 1, gnx, gny, lx, ly

        ! we print the base state into the file
        write(77) vbx(0:nx-1,0:ny-1),vby(0:nx-1,0:ny-1),vbz(0:nx-1,0:ny-1)
        write(77) wbx(0:nx-1,0:ny-1),wby(0:nx-1,0:ny-1),wbz(0:nx-1,0:ny-1)

        close(77)

    end subroutine


! ############################################################
!
!    Basic output
!
! ############################################################
!**********************************************************************
    subroutine output1(svx,svy,svz,srho)
!*************************************************************************
    !****f* output/output1
    !  NAME
    !    output1
    !
    !  DESCRIPTION
    !    Print the elapsed time in the time stepping loop and the expected remaining time.
    !    Print the quadratic energy and the growhrate of the velocity field.
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields in spectral space
    !***

        implicit none

        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(in) :: svx,svy,svz,srho
        double precision :: quad_velo, growthrate
        double precision, save :: quad_velo_old
        double precision, save :: time_start=0., time_old, time_current, time_loop
        double precision :: trash

        ! only to avoid a compiler warning "srho is not used" while keeping an identical list of arguments for output subroutine
        ! this is really a "cosmetic" change...
        trash = srho(1,0,0,0)

        ! we calculate the time already spent in the time stepping loop and the expected remaining time
        if (time_start==0) then
            call cpu_time(time_start)
            time_current = time_start
            time_old = time_start
            time_loop = 0.
        else
            call cpu_time(time_current)
            time_loop = time_current - time_old
            time_old = time_current
        end if

        ! we determine the mean quadratic velocity and the growthrate between two successive calls to the output1 subroutine
        quad_velo = sum(svx(:,0:nk1-1,0:nk2-1,0:nk3-1)**2 + svy(:,0:nk1-1,0:nk2-1,0:nk3-1)**2 + svz(:,0:nk1-1,0:nk2-1,0:nk3-1)**2)

#ifdef MPI
        call MPI_allreduce(MPI_IN_PLACE,quad_velo,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
#endif


#ifdef MPI
        if (rank==0) then
#endif
            if (quad_velo_old>epsilo) then
                growthrate=0.5D0*log(quad_velo/quad_velo_old)/(output1_period*dt)
            else
                growthrate=0.
            endif

            write(*,*)
            write(*,'(A4,TR2,I6,TR3,A6,TR2,F10.3)') 'it =', it, 'time =', it*dt + begin
            write(*,'(A40,TR2,I10,TR3,I10)') '   cpu time in sec (elapsed/remaining) =',int(time_current-time_start), &
                                              int(time_loop/output1_period*(itmax-it))

            write(*,*) '  mean quadratic velocity =', quad_velo
            write(*,*) '  growthrate =', growthrate
#ifdef MPI
        end if
#endif

        quad_velo_old = quad_velo

    end subroutine




! ############################################################
!
!    Data file output
!
! ############################################################
#ifdef MPI
! *************************************************************************
    subroutine output2(svx,svy,svz,srho,wkx,wky,wkz,wkrho)
! *************************************************************************
    !****f* output/output2
    !  NAME
    !    output2
    !
    !  DESCRIPTION
    !    Record in a binary file the velocity, density and vorticity fields expressed in physical space.
    !    This version uses MPI-IO (MPI Input-Output) library. The output format is exactly the same as for
    !    the sequential version (see below).
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields in spectral space
    !
    !  OUPUTS  
    !    wkx,wky,wkz,wkrho: working arrays
    !***
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(in) :: svx,svy,svz,srho
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx,wky,wkz,wkrho
        integer :: descriptor
        integer, dimension(MPI_STATUS_SIZE) :: mpi_status
        character(len=50) :: file_name
        character(len=10) :: sufx

        call time2string(it*dt + begin,sufx)
        file_name='velo_rho_vort.'//sufx

        call MPI_file_open(MPI_COMM_WORLD,file_name,MPI_MODE_WRONLY+MPI_MODE_CREATE,MPI_INFO_NULL,descriptor,code)

        if (rank==0) then

            ! we first write meta-data in order to help the post-treatment of the data file and avoid errors
            ! Fortran adds before and after every recording an integer that corresponds to the sixe of the recording
            ! expressed in bytes. We must also take into account these extra integers while post-treating the file.

            ! the exact format file is (i and j are integers corresponding to the preceding or following recording):

            ! i,1,gnx,gny,gnz,lx,ly,lz,de_aliasing,type_trunc,rtrunc_x,rtrunc_y,rtrunc_z,nu,stratification,
            ! xns,schm,omega2,perturbative,linear,j,i,pvx,j,i,pvy,j,i,pvz,j,i,prho,j,i,pvortx,j,i,pvorty,j,i,pvortz,j,
            ! i,vbx,vby,vbz,j,i,wbx,wby,wbz,j

            ! Contrary to a sequential code, MPI-IO does not add automatically the size of a record at the begining and the end
            ! We add it here explicitely
            call MPI_file_write_shared(descriptor,124,              1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,1,                1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,gnx,              1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,gny,              1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,gnz,              1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,lx,               1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,ly,               1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,lz,               1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,dt,               1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,de_aliasing,      1, MPI_LOGICAL          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,type_trunc,       1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,rtrunc_x,         1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,rtrunc_y,         1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,rtrunc_z,         1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,nu,               1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,stratification,   1, MPI_LOGICAL          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,xns,              1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,schm,             1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,omega2,           1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,perturbative,     1, MPI_LOGICAL          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,linear,           1, MPI_LOGICAL          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,124,              1, MPI_INTEGER          ,mpi_status,code)

            call MPI_file_write_shared(descriptor,8,                1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,it*dt+begin,      1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,8,                1, MPI_INTEGER          ,mpi_status,code)
        end if

        ! we switch the velocity into physical space and we print it in the file
        wkx = svx
        wky = svy
        wkz = svz
        wkrho = srho

        call bck_fft(wkx,fft_work)
        call bck_fft(wky,fft_work)
        call bck_fft(wkz,fft_work)
        if (stratification) then
            call bck_fft(wkrho,fft_work)
        end if

        call write_array(wkx,descriptor)
        call write_array(wky,descriptor)
        call write_array(wkz,descriptor)
        call write_array(wkrho,descriptor)

        ! we determine the vorticity, we switch it into physical space and we print it in the file
        call curl(svx,svy,svz,wkx,wky,wkz)

        call bck_fft(wkx,fft_work)
        call bck_fft(wky,fft_work)
        call bck_fft(wkz,fft_work)

        call write_array(wkx,descriptor)
        call write_array(wky,descriptor)
        call write_array(wkz,descriptor)

        if (rank==0) then
            call MPI_file_write_shared(descriptor,3*8*gnx*gny,1, MPI_INTEGER,mpi_status,code)
            call MPI_file_write_shared(descriptor,vbx(0:nx-1,0:ny-1),nx*ny,MPI_DOUBLE_PRECISION,mpi_status,code)
            call MPI_file_write_shared(descriptor,vby(0:nx-1,0:ny-1),nx*ny,MPI_DOUBLE_PRECISION,mpi_status,code)
            call MPI_file_write_shared(descriptor,vbz(0:nx-1,0:ny-1),nx*ny,MPI_DOUBLE_PRECISION,mpi_status,code)
            call MPI_file_write_shared(descriptor,3*8*gnx*gny,1, MPI_INTEGER,mpi_status,code)

            call MPI_file_write_shared(descriptor,3*8*gnx*gny,1, MPI_INTEGER,mpi_status,code)
            call MPI_file_write_shared(descriptor,wbx(0:nx-1,0:ny-1),nx*ny,MPI_DOUBLE_PRECISION,mpi_status,code)
            call MPI_file_write_shared(descriptor,wby(0:nx-1,0:ny-1),nx*ny,MPI_DOUBLE_PRECISION,mpi_status,code)
            call MPI_file_write_shared(descriptor,wbz(0:nx-1,0:ny-1),nx*ny,MPI_DOUBLE_PRECISION,mpi_status,code)
            call MPI_file_write_shared(descriptor,3*8*gnx*gny,1, MPI_INTEGER,mpi_status,code)
        end if

        call MPI_file_close(descriptor,code)

    end subroutine

! ***********************************************************
    subroutine write_array(pa,descriptor_file)
! ***********************************************************
    !****f* output/write_array
    !  NAME
    !    write_array
    !
    !  DESCRIPTION
    !    When running in MPI, this subroutine allows to write an array that is splitted between
    !    different processes in a single data file.
    !
    !  INPUTS
    !    pa: field to record in physical space
    !    descriptor_file: MPI descriptor of the file where the field should be recorded
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1),intent(in) :: pa
        integer, intent(in) :: descriptor_file
        integer, dimension(MPI_STATUS_SIZE) :: mpi_status

        integer :: ix, iy, iz

        do iz=0,nz-1
            do iy=0,ny-1
                do ix=0,nx-1
                    wk_io(ix,iy,iz) = pa(ix,iy,iz)
                end do
            end do
        end do

        if (rank==0) then
            call MPI_file_write_shared(descriptor_file,8*gnx*gny*gnz,1, MPI_INTEGER,mpi_status,code)
        end if

        call MPI_file_write_ordered(descriptor_file,wk_io,nx*ny*nz,MPI_DOUBLE_PRECISION,mpi_status,code)

        if (rank==0) then
            call MPI_file_write_shared(descriptor_file,8*gnx*gny*gnz,1, MPI_INTEGER,mpi_status,code)
        end if

    end subroutine

#else
! *************************************************************************
    subroutine output2(svx,svy,svz,srho,wkx,wky,wkz,wkrho)
! *************************************************************************
    !****f* output/output2
    !  NAME
    !    output2
    !
    !  DESCRIPTION
    !    Record in a binary file the velocity, density and vorticity fields expressed in physical space.
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields in spectral space
    !
    !  OUPUTS  
    !    wkx,wky,wkz,wkrho: working arrays
    !***
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(in) :: svx,svy,svz,srho
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx,wky,wkz,wkrho
        character(len=50) :: file_name
        character(len=10) :: sufx

        ! we first write meta-data in order to help the post-treatment of the data file and avoid errors
        ! Fortran adds before and after every recording an integer that corresponds to the size of the recording
        ! expressed in bytes. We must also take into account these extra integers while post-treating the file.

        ! the exact format file is (i and j are integers corresponding to the preceding or following recording):

        ! i,1,gnx,gny,gnz,lx,ly,lz,de_aliasing,type_trunc,rtrunc_x,rtrunc_y,rtrunc_z,nu,stratification,
        ! xns,schm,omega2,perturbative,linear,j,i,pvx,j,i,pvy,j,i,pvz,j,i,prho,j,i,pvortx,j,i,pvorty,j,i,pvortz,j,
        ! i,vbx,vby,vbz,j,i,wbx,wby,wbz,j

        call time2string(it*dt + begin,sufx)
        file_name='velo_rho_vort.'//sufx

        open (76,file=file_name,form='unformatted')

        write(76) 1, nx, ny, nz, lx, ly, lz, dt, de_aliasing, type_trunc, rtrunc_x, rtrunc_y, rtrunc_z, &
                  nu, stratification, xns, schm, omega2, perturbative, linear

        write(76) it*dt + begin

        ! we switch the velocity into physical space and we print it in the file
        wkx = svx
        wky = svy
        wkz = svz
        wkrho = srho

        call bck_fft(wkx,fft_work)
        call bck_fft(wky,fft_work)
        call bck_fft(wkz,fft_work)
        if (stratification) then
            call bck_fft(wkrho,fft_work)
        end if

        call write_array(wkx,76)
        call write_array(wky,76)
        call write_array(wkz,76)
        call write_array(wkrho,76)

        ! we determine the vorticity, we switch it into physical space and we print it in the file
        call curl(svx,svy,svz,wkx,wky,wkz)

        call bck_fft(wkx,fft_work)
        call bck_fft(wky,fft_work)
        call bck_fft(wkz,fft_work)

        call write_array(wkx,76)
        call write_array(wky,76)
        call write_array(wkz,76)

        ! we print the base state into the file
        write(76) vbx(0:nx-1,0:ny-1),vby(0:nx-1,0:ny-1),vbz(0:nx-1,0:ny-1)
        write(76) wbx(0:nx-1,0:ny-1),wby(0:nx-1,0:ny-1),wbz(0:nx-1,0:ny-1)

        close(76)

    end subroutine

! ***********************************************************
    subroutine write_array(pa,id_file)
! ***********************************************************
    !****f* output/write_array
    !  NAME
    !    write_array
    !
    !  DESCRIPTION
    !    Allow to write a subsection of an array into a file without using too much stack memory which
    !    can be an issue on some systems with limited stack memory.
    !    Indeed, surprisingly the following instructions are responsible for the use of an amount of
    !    stack memory corresponding to the **sum** of all the recorded arrays:
    !        write(1) aax(0:nx-1,0:ny-1,0:nz-1)
    !        write(1) aay(0:nx-1,0:ny-1,0:nz-1)
    !        (...)
    !
    !    This subroutine makes use of a single storage array that is static and so defined in data memory.
    !
    !  INPUTS
    !    pa: field to record in physical space
    !    id_file: identifier of the file where the field should be recorded
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1),intent(in) :: pa
        integer, intent(in) :: id_file
        integer :: ix, iy, iz

        do iz=0,nz-1
            do iy=0,ny-1
                do ix=0,nx-1
                    wk_io(ix,iy,iz) = pa(ix,iy,iz)
                end do
            end do
        end do

        write(id_file) wk_io

    end subroutine
#endif



! *************************************************************************
    subroutine output3(svx,svy,svz,srho,wkx,wky,wkz,wkrho)
! *************************************************************************
    !****f* output/output3
    !  NAME
    !    output3
    !
    !  DESCRIPTION
    !    Does nothing. The user can write his own subroutine according to his needs.
    !    Be aware to write a MPI version of this output3 subroutine if required.
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields in spectral space
    !
    !  OUPUTS  
    !    wkx,wky,wkz,wkrho: working arrays
    !***
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(in) :: svx,svy,svz,srho
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx,wky,wkz,wkrho
				double precision :: trash_dbl
				
				! Only to avoid the warning "The variable 'xxx' was not used" that is issued by some compilers.
        trash_dbl = svx(1,0,0,0)
				trash_dbl = svy(1,0,0,0)
				trash_dbl = svz(1,0,0,0)
				trash_dbl = srho(1,0,0,0)
				wkx(1,0,0,0) = trash_dbl
				wky(1,0,0,0) = trash_dbl
				wkz(1,0,0,0) = trash_dbl
				wkrho(1,0,0,0) = trash_dbl
				
    end subroutine

end module
